"""Add task result

Revision ID: 05cf96d6fcae
Revises: 30d6f6636351
Create Date: 2020-03-20 19:16:48.520652

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "05cf96d6fcae"
down_revision = "30d6f6636351"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("tasks", sa.Column("result", sa.PickleType(), nullable=True))


def downgrade():
    op.drop_column("tasks", "result")



"""Add is_hook on Task

Revision ID: 9817ccf13cb5
Revises: 46e4acde004e
Create Date: 2022-10-21 15:32:26.186650

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "9817ccf13cb5"
down_revision = "46e4acde004e"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("tasks", sa.Column("is_hook", sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column("tasks", "is_hook")



def test_view(client):
    resp = client.get("/")
    assert resp.status_code == 200
    assert not resp.is_json
    assert resp.mimetype == "text/html"
    assert "<title>Celery Director</title>" in str(resp.data)
    assert 'const REPO_LINK="https://github.com/ovh/celery-director"' in str(resp.data)


def test_404_view(client):
    resp = client.get("/notfound")
    assert resp.status_code == 404
    assert not resp.is_json
    assert resp.mimetype == "text/html"
    assert "<title>404 Not Found</title>" in str(resp.data)


def test_api(client):
    resp = client.get("/api/ping")
    assert resp.status_code == 200
    assert resp.is_json
    assert resp.mimetype == "application/json"
    assert resp.json == {"message": "pong"}


def test_404_api(client):
    resp = client.get("/api/notfound")
    assert resp.status_code == 404
    assert resp.is_json
    assert resp.mimetype == "application/json"
    assert "The requested URL was not found" in resp.json["error"]



from flask_json_schema import JsonValidationError
from jsonschema.validators import validator_for
from celery.schedules import crontab

from director.exceptions import WorkflowSyntaxError


def validate(payload, schema):
    """Validate a payload according to a given schema"""
    validator_cls = validator_for(schema)
    validator = validator_cls(schema=schema)
    errors = list(validator.iter_errors(payload))
    if errors:
        raise JsonValidationError("Payload is not valid", errors)


def format_schema_errors(e):
    """Format FlaskJsonSchema validation errors"""
    return {
        "error": e.message,
        "errors": [validation_err.message for validation_err in e.errors],
    }


def build_celery_schedule(workflow_name, data):
    """A celery schedule can accept seconds or crontab"""

    def _handle_schedule(schedule):
        try:
            value = float(schedule)
        except ValueError:
            m, h, dw, dm, my = schedule.split(" ")
            value = crontab(
                minute=m,
                hour=h,
                day_of_month=dm,
                month_of_year=my,
                day_of_week=dw,
            )
        return value

    def _handle_crontab(ct):
        m, h, dm, my, dw = ct.split(" ")
        return crontab(
            minute=m,
            hour=h,
            day_of_month=dm,
            month_of_year=my,
            day_of_week=dw,
        )

    excluded_keys = ["payload"]
    keys = [k for k in data.keys() if k not in excluded_keys]

    schedule_functions = {
        # Legacy syntax for backward compatibility
        "schedule": _handle_schedule,
        # Current syntax
        "crontab": _handle_crontab,
        "interval": float,
    }

    if len(keys) != 1 or keys[0] not in schedule_functions.keys():
        # When there is no key (schedule, interval, crontab) in the periodic configuration
        raise WorkflowSyntaxError(workflow_name)

    schedule_key = keys[0]
    schedule_input = data[schedule_key]
    try:
        # Apply the function mapped to the schedule type
        return str(schedule_input), schedule_functions[schedule_key](schedule_input)
    except Exception:
        raise WorkflowSyntaxError(workflow_name)



from datetime import datetime, timedelta
from distutils.util import strtobool

import pytz
from flask import abort, jsonify, request
from flask import current_app as app

from director.api import api_bp, validate
from director.auth import auth
from director.builder import WorkflowBuilder
from director.exceptions import WorkflowNotFound
from director.extensions import cel_workflows, schema
from director.models.workflows import Workflow
from director.utils import build_celery_schedule


def _get_workflow(workflow_id):
    workflow = Workflow.query.filter_by(id=workflow_id).first()
    if not workflow:
        abort(404, f"Workflow {workflow_id} not found")
    return workflow


def _execute_workflow(project, name, payload={}):
    fullname = f"{project}.{name}"

    # Check if the workflow exists
    try:
        wf = cel_workflows.get_by_name(fullname)
        if "schema" in wf:
            validate(payload, wf["schema"])
    except WorkflowNotFound:
        abort(404, f"Workflow {fullname} not found")

    # Create the workflow in DB
    obj = Workflow(project=project, name=name, payload=payload)
    obj.save()

    # Build the workflow and execute it
    data = obj.to_dict()
    workflow = WorkflowBuilder(obj.id)
    workflow.run()

    app.logger.info(f"Workflow sent : {workflow.canvas}")
    return obj.to_dict(), workflow


def _cancel_workflow(obj):
    workflow = WorkflowBuilder(obj.id)
    workflow.cancel()

    app.logger.info(f"Workflow {obj.id} canceled")
    return obj.to_dict(), workflow


@api_bp.route("/workflows", methods=["POST"])
@auth.login_required
@schema.validate(
    {
        "required": ["project", "name", "payload"],
        "additionalProperties": False,
        "properties": {
            "project": {"type": "string"},
            "name": {"type": "string"},
            "payload": {"type": "object"},
        },
    }
)
def create_workflow():
    project, name, payload = (
        request.get_json()["project"],
        request.get_json()["name"],
        request.get_json()["payload"],
    )
    data, _ = _execute_workflow(project, name, payload)
    return jsonify(data), 201


@api_bp.route("/workflows/<workflow_id>/relaunch", methods=["POST"])
@auth.login_required
def relaunch_workflow(workflow_id):
    obj = _get_workflow(workflow_id)
    data, _ = _execute_workflow(obj.project, obj.name, obj.payload)
    return jsonify(data), 201


@api_bp.route("/workflows/<workflow_id>/cancel", methods=["POST"])
@auth.login_required
def cancel_workflow(workflow_id):
    obj = _get_workflow(workflow_id)
    data, _ = _cancel_workflow(obj)
    return jsonify(data), 201


@api_bp.route("/workflows")
@auth.login_required
def list_workflows():
    page = request.args.get("page", type=int, default=1)
    # Convert with_payload arg to boolean, if we encounter an error,
    # ignore and set with_payload to its default value (True)
    # We don't do request.args.get with type=bool because it doesn't seem to work as expected because
    # this most likely casts the string as bool which is not the proper string to bool conversion we need
    try:
        with_payload = strtobool(
            request.args.get("with_payload", type=str, default="True")
        )
    except ValueError:
        with_payload = True
    per_page = request.args.get(
        "per_page", type=int, default=app.config["WORKFLOWS_PER_PAGE"]
    )
    workflows = Workflow.query.order_by(Workflow.created_at.desc()).paginate(
        page, per_page
    )
    return jsonify([w.to_dict(with_payload=with_payload) for w in workflows.items])


@api_bp.route("/workflows/<workflow_id>")
@auth.login_required
def get_workflow(workflow_id):
    workflow = _get_workflow(workflow_id)
    tasks = [t.to_dict() for t in workflow.tasks]

    resp = workflow.to_dict()
    resp.update({"tasks": tasks})
    return jsonify(resp)


@api_bp.route("/definitions")
@auth.login_required
def list_definitions():
    workflow_definitions = []
    for fullname, definition in sorted(cel_workflows.workflows.items()):
        project, name = fullname.split(".", 1)
        workflow_definitions.append(
            {"fullname": fullname, "project": project, "name": name, **definition}
        )
    return jsonify(workflow_definitions)



"""Add cascade in task model for workflow deletion

Revision ID: 46e4acde004e
Revises: 2ac615d6850b
Create Date: 2021-07-16 11:54:27.946028

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = "46e4acde004e"
down_revision = "2ac615d6850b"
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        # SQLite does not support to alter constraints on existing tables.
        # Batch mode is used to copy data to a temporary table meanwhile creating a brand new
        # table with the required constraint.
        with op.batch_alter_table("tasks") as batch_op:
            batch_op.drop_constraint(
                op.f("fk_tasks_workflow_id_workflows"), type_="foreignkey"
            )
            batch_op.create_foreign_key(
                op.f("fk_tasks_workflow_id_workflows"),
                "workflows",
                ["workflow_id"],
                ["id"],
                ondelete="cascade",
            )
        return

    op.drop_constraint(
        op.f("fk_tasks_workflow_id_workflows"), "tasks", type_="foreignkey"
    )
    op.create_foreign_key(
        op.f("fk_tasks_workflow_id_workflows"),
        "tasks",
        "workflows",
        ["workflow_id"],
        ["id"],
        ondelete="cascade",
    )


def downgrade():
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        with op.batch_alter_table("tasks") as batch_op:
            batch_op.drop_constraint(
                op.f("fk_tasks_workflow_id_workflows"), type_="foreignkey"
            )
            batch_op.create_foreign_key(
                op.f("fk_tasks_workflow_id_workflows"),
                "workflows",
                ["workflow_id"],
                ["id"],
            )
        return

    op.drop_constraint(
        op.f("fk_tasks_workflow_id_workflows"), "tasks", type_="foreignkey"
    )
    op.create_foreign_key(
        op.f("fk_tasks_workflow_id_workflows"),
        "tasks",
        "workflows",
        ["workflow_id"],
        ["id"],
    )



from werkzeug.security import generate_password_hash
from terminaltables import AsciiTable


from director.context import pass_ctx
from director.exceptions import UserNotFound
from director.models.users import User

import click


def _get_users():
    return User.query.all()


@click.group()
def user():
    """Manage the users"""


@user.command(name="list")
@pass_ctx
def list_users(ctx):
    """Display users"""
    data = [["Username"]]
    users = _get_users()

    for user in users:
        data.append([user.username])

    table = AsciiTable(data)
    table.inner_row_border = True
    click.echo(table.table)


@user.command(name="create")
@click.argument("username")
@click.password_option()
@pass_ctx
def create_user(ctx, username, password):
    """Create user"""
    user = User(username=username, password=generate_password_hash(password))
    user.save()


@user.command(name="update")
@click.argument("username")
@click.password_option()
@pass_ctx
def update_user(ctx, username, password):
    """Update user"""
    user = User(username=username, password=generate_password_hash(password))
    try:
        user.update()
    except UserNotFound as e:
        click.echo(str(e))


@user.command(name="delete")
@click.argument("username")
@pass_ctx
def delete_user(ctx, username):
    """delete user"""
    user = User.query.filter_by(username=username).first()
    if not user:
        click.echo(f"User {username} not found")
        return

    user.delete()



from director.extensions import DirectorSentry, cel


def test_sentry_enrich_data(app, create_builder):
    _, wf = create_builder("example", "WORKFLOW", {})

    sentry = DirectorSentry()
    sentry.init_app(app)
    tags = sentry.enrich_tags(
        {"foo": "bar"}, wf.workflow_id, cel.tasks.get("TASK_EXAMPLE")
    )
    assert tags == {
        "foo": "bar",
        "celery_task_name": "TASK_EXAMPLE",
        "director_workflow_id": str(wf.workflow_id),
        "director_workflow_project": "example",
        "director_workflow_name": "example.WORKFLOW",
    }

    extra = sentry.enrich_extra(
        {"foo": "bar"}, [{"key": "value"}], {"payload": {"hello": "world"}}
    )
    assert extra == {
        "foo": "bar",
        "task-args": [{"key": "value"}],
        "workflow-payload": {"hello": "world"},
    }



from sqlalchemy_utils import JSONType

try:
    from sqlalchemy.dialects.postgresql import JSONB

    has_postgres_jsonb = True
except ImportError:
    has_postgres_jsonb = False


class JSONBType(JSONType):
    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql" and has_postgres_jsonb:
            return dialect.type_descriptor(JSONB())
        return super(JSONBType, self).load_dialect_impl(dialect)



from flask import current_app as app
from flask import jsonify, abort
from flask_httpauth import HTTPBasicAuth

from director.models.users import User

from werkzeug.security import check_password_hash

ANONYMOUS_USERNAME = "anonymous"

auth = HTTPBasicAuth()


@auth.verify_password
def verify_password(username, password):
    # no need to check if auth is disabled
    if not app.config.get("AUTH_ENABLED"):
        return ANONYMOUS_USERNAME

    # return false if auth is enabled and no username is set
    if not username:
        return

    user = User.query.filter_by(username=username).first()
    if not user:
        return

    # auth is enabled, needs to check the credz
    if check_password_hash(user.password, password):
        return username


@auth.error_handler
def unauthorized():
    response = jsonify(
        {
            "status": 401,
            "error": "unauthorized",
            "message": "Please authenticate to access this API.",
        }
    )
    response.status_code = 401
    return response



import logging

from sqlalchemy.orm import load_only

from director.builder import WorkflowBuilder
from director.extensions import cel, db
from director.models.workflows import Workflow

logger = logging.getLogger()


@cel.task()
def execute(workflow, payload):
    project, name = workflow.split(".")
    c_obj = Workflow(project=project, name=name, payload=payload, periodic=True)
    c_obj.save()

    # Build the workflow and execute it
    workflow = WorkflowBuilder(c_obj.id)
    workflow.run()

    c_obj_dict = c_obj.to_dict()

    # Force commit before ending the function to ensure the ongoing transaction
    # does not end up in a "idle in transaction" state on PostgreSQL
    c_obj.commit()

    return c_obj_dict


@cel.task()
def cleanup(retentions):
    count = 0
    for workflow_name, retention in retentions.items():
        project, name = workflow_name.split(".")
        logger.info(f"Cleaning {workflow_name} (retention of {retention})")

        bind = db.session.get_bind()
        if bind.engine.name == "sqlite":
            # SQLite does not use ON DELETE CASCADE by default
            db.session.execute("PRAGMA foreign_keys=ON")

        workflows = (
            db.session.query(Workflow)
            .options(load_only(Workflow.id))
            .filter_by(project=project, name=name)
            .order_by(Workflow.created_at.desc())
            .offset(retention)
            .all()
        )

        ids = [workflow.id for workflow in workflows]
        if not ids:
            logger.info(f"No need to clean {workflow_name}")
            continue

        Workflow.query.filter(Workflow.id.in_(ids)).delete(synchronize_session=False)
        db.session.commit()
        count += len(ids)
        logger.info(f"Deleted workflows: {len(ids)}")
    return count



from datetime import datetime, timedelta
from unittest.mock import patch

from director.extensions import db
from director.models.workflows import Workflow


DEFAULT_PAYLOAD = {"project": "example", "name": "WORKFLOW", "payload": {}}


def test_pong(client):
    resp = client.get("/api/ping")
    assert resp.status_code == 200
    assert resp.json == {"message": "pong"}


def test_post_workflow_without_payload(client, no_worker):
    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    assert resp.status_code == 201
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {},
        "periodic": False,
        "project": "example",
        "status": "pending",
    }


def test_post_workflow_with_payload(client, no_worker):
    payload = {**DEFAULT_PAYLOAD, "payload": {"nested": {"foo": "bar"}}}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 201
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {"nested": {"foo": "bar"}},
        "periodic": False,
        "project": "example",
        "status": "pending",
    }


def test_post_workflow_required_props(client):
    resp = client.post("/api/workflows", json={})
    assert resp.status_code == 400
    assert resp.json["error"] == "Error validating against schema"
    assert len(resp.json["errors"]) == 3
    for prop in ["project", "name", "payload"]:
        assert f"{prop}" in " ".join(resp.json["errors"])

    resp = client.post("/api/workflows", json={"project": "example"})
    assert resp.status_code == 400
    assert len(resp.json["errors"]) == 2
    for prop in ["name", "payload"]:
        assert f"{prop}" in " ".join(resp.json["errors"])

    resp = client.post(
        "/api/workflows", json={"project": "example", "name": "WORKFLOW"}
    )
    assert resp.status_code == 400
    assert len(resp.json["errors"]) == 1
    assert "payload" in resp.json["errors"][0]


def test_post_workflow_unexpected_properties(client):
    payload = {**DEFAULT_PAYLOAD, "foo": "bar"}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 400
    assert resp.json["error"] == "Error validating against schema"
    assert "'foo' was unexpected" in resp.json["errors"][0]


def test_relaunch_not_existing_workflow(client):
    resp = client.post(
        "/api/workflows/280d0c4065d04063a800a3b674562711/relaunch", json={}
    )
    assert resp.status_code == 404


def test_relaunch_workflow(client, no_worker):
    payload = {**DEFAULT_PAYLOAD, "payload": {"nested": {"foo": "bar"}}}
    resp = client.post("/api/workflows", json=payload)
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflow_id = resp.json["id"]
    assert resp.status_code == 201

    resp = client.post(f"/api/workflows/{workflow_id}/relaunch", json={})
    assert resp.status_code == 201
    assert resp.json == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {"nested": {"foo": "bar"}},
        "periodic": False,
        "project": "example",
        "status": "pending",
    }

    # workflows are the same
    resp = client.get("/api/workflows")
    assert len(resp.json) == 2

    assert resp.json[0] == resp.json[1]

    # tasks are the same
    resp = client.get("/api/workflows")
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflows = resp.json
    tasks1 = [
        t["key"]
        for t in client.get(f"/api/workflows/{workflows[0]['id']}").json.get("tasks")
    ]
    tasks2 = [
        t["key"]
        for t in client.get(f"/api/workflows/{workflows[1]['id']}").json.get("tasks")
    ]
    assert tasks1 == tasks2


def test_cancel_not_existing_workflow(client, no_worker):
    resp = client.post(
        "/api/workflows/280d0c4065d04063a800a3b674562711/cancel", json={}
    )
    assert resp.status_code == 404


def test_cancel_workflow(client, no_worker):
    resp = client.post(
        "/api/workflows",
        json={"project": "example", "name": "DELAY_TASK", "payload": {}},
    )
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflow_id = resp.json["id"]
    assert resp.status_code == 201

    resp = client.post(f"/api/workflows/{workflow_id}/cancel", json={})
    assert resp.status_code == 201
    assert resp.json["status"] == "canceled"

    tasks = [
        t["status"]
        for t in client.get(f"/api/workflows/{workflow_id}").json.get("tasks")
    ]
    assert tasks[0] == "canceled"


def test_not_found_workflows(client, no_worker):
    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    assert resp.status_code == 201

    payload = {**DEFAULT_PAYLOAD, "name": "NOT_FOUND"}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 404
    assert resp.json == {"error": "Workflow example.NOT_FOUND not found"}

    payload = {**DEFAULT_PAYLOAD, "project": "NOT_FOUND"}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 404
    assert resp.json == {"error": "Workflow NOT_FOUND.WORKFLOW not found"}


def test_list_workflows(client, no_worker):
    resp = client.get("/api/workflows")
    assert resp.status_code == 200
    assert resp.json == []

    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    assert resp.status_code == 201

    resp = client.get("/api/workflows")
    assert resp.json == [
        {
            "fullname": "example.WORKFLOW",
            "name": "WORKFLOW",
            "payload": {},
            "periodic": False,
            "project": "example",
            "status": "pending",
        }
    ]

    for i in range(9):
        client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    resp = client.get("/api/workflows")
    assert len(resp.json) == 10


def test_list_paginated_workflows(app, client, no_worker):
    # SQLite db.Datetime() does not include microseconds, so
    # we need to fix dates to be able to sort workflows.
    now = datetime.now()
    with app.app_context():
        for i in range(20):
            date = now + timedelta(minutes=i)
            w = Workflow(
                created_at=date,
                updated_at=date,
                project="example",
                name="WORKFLOW",
                payload={"i": i + 1},
            )
            db.session.add(w)
        db.session.commit()

    # DIRECTOR_WORKFLOWS_PER_PAGE=15 in tests/workflows/.env
    resp = client.get("/api/workflows")
    assert len(resp.json) == 15

    # Last item
    resp = client.get("/api/workflows?per_page=1")
    assert len(resp.json) == 1
    assert resp.json[0]["payload"]["i"] == 20

    # Last X items
    resp = client.get("/api/workflows?per_page=4&page=1")
    assert len(resp.json) == 4
    assert [w["payload"]["i"] for w in resp.json] == [20, 19, 18, 17]

    # Last X items, second page
    resp = client.get("/api/workflows?per_page=4&page=2")
    assert len(resp.json) == 4
    assert [w["payload"]["i"] for w in resp.json] == [16, 15, 14, 13]

    # Remaining items
    resp = client.get("/api/workflows?per_page=3&page=7")
    assert len(resp.json) == 2
    assert [w["payload"]["i"] for w in resp.json] == [2, 1]

    # Out of range
    resp = client.get("/api/workflows?per_page=4&page=6")
    assert resp.status_code == 404


def test_get_workflow(client, no_worker):
    resp = client.post("/api/workflows", json=DEFAULT_PAYLOAD)
    with patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[]):
        workflow_id = resp.json["id"]

    resp = client.get(f"/api/workflows/{workflow_id}")
    assert resp.status_code == 200
    assert len(resp.json["tasks"]) == 1

    resp_json_copy = resp.json.copy()
    del resp_json_copy["tasks"][0]["task"]

    assert resp_json_copy == {
        "fullname": "example.WORKFLOW",
        "name": "WORKFLOW",
        "payload": {},
        "periodic": False,
        "project": "example",
        "status": "pending",
        "tasks": [
            {
                "key": "TASK_EXAMPLE",
                "previous": [],
                "result": None,
                "status": "pending",
                "is_hook": False,
            }
        ],
    }


@patch("tests.conftest.DirectorResponse._KEYS_TO_REMOVE", new=[])
def test_get_workflow_simple_chain(client, no_worker):
    payload = {**DEFAULT_PAYLOAD, "name": "SIMPLE_CHAIN"}
    workflow_id = client.post("/api/workflows", json=payload).json["id"]

    resp = client.get(f"/api/workflows/{workflow_id}")
    assert resp.status_code == 200
    assert len(resp.json["tasks"]) == 3


def test_get_not_found_workflow(client):
    resp = client.get("/api/workflows/e0ee42d8-e00f-4557-8f75-781106dbbea4")
    assert resp.status_code == 404


def test_schema(client, no_worker):
    payload = {"project": "schemas", "name": "SIMPLE_SCHEMA", "payload": {}}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 400
    assert resp.json == {
        "error": "Payload is not valid",
        "errors": ["'name' is a required property"],
    }

    payload["payload"]["price"] = "foo"
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 400
    assert resp.json["error"] == "Payload is not valid"
    assert sorted(resp.json["errors"]) == sorted(
        ["'name' is a required property", "'foo' is not of type 'number'"]
    )

    payload["payload"] = {"name": "foo", "price": 100}
    resp = client.post("/api/workflows", json=payload)
    assert resp.status_code == 201


def test_list_definitions(client, no_worker):
    resp = client.get("/api/definitions")
    assert resp.status_code == 200
    assert resp.json == [
        {
            "fullname": "example.CELERY_ERROR_MULTIPLE_TASKS",
            "name": "CELERY_ERROR_MULTIPLE_TASKS",
            "project": "example",
            "tasks": ["TASK_A", "TASK_CELERY_ERROR"],
        },
        {
            "fullname": "example.CELERY_ERROR_ONE_TASK",
            "name": "CELERY_ERROR_ONE_TASK",
            "project": "example",
            "tasks": ["TASK_CELERY_ERROR"],
        },
        {
            "fullname": "example.DELAY_TASK",
            "name": "DELAY_TASK",
            "project": "example",
            "tasks": ["DELAY"],
        },
        {
            "fullname": "example.ERROR",
            "name": "ERROR",
            "project": "example",
            "tasks": ["TASK_ERROR"],
        },
        {
            "fullname": "example.FAILURE_HOOK",
            "hooks": {"failure": "TASK_B"},
            "name": "FAILURE_HOOK",
            "project": "example",
            "tasks": ["TASK_A", "TASK_ERROR"],
        },
        {
            "fullname": "example.RETURN_EXCEPTION",
            "name": "RETURN_EXCEPTION",
            "project": "example",
            "tasks": ["STR", "TASK_ERROR"],
        },
        {
            "fullname": "example.RETURN_VALUES",
            "name": "RETURN_VALUES",
            "project": "example",
            "tasks": ["STR", "INT", "LIST", "NONE", "DICT", "NESTED"],
        },
        {
            "fullname": "example.SIMPLE_CHAIN",
            "name": "SIMPLE_CHAIN",
            "project": "example",
            "tasks": ["TASK_A", "TASK_B", "TASK_C"],
        },
        {
            "fullname": "example.SIMPLE_CHAIN_ERROR",
            "name": "SIMPLE_CHAIN_ERROR",
            "project": "example",
            "tasks": ["TASK_A", "TASK_B", "TASK_ERROR"],
        },
        {
            "fullname": "example.SIMPLE_GROUP",
            "name": "SIMPLE_GROUP",
            "project": "example",
            "tasks": [
                "TASK_A",
                {"EXAMPLE_GROUP": {"tasks": ["TASK_B", "TASK_C"], "type": "group"}},
            ],
        },
        {
            "fullname": "example.SIMPLE_GROUP_ERROR",
            "name": "SIMPLE_GROUP_ERROR",
            "project": "example",
            "tasks": [
                "TASK_A",
                {"EXAMPLE_GROUP": {"tasks": ["TASK_ERROR", "TASK_C"], "type": "group"}},
            ],
        },
        {
            "fullname": "example.SUCCESSANDFAILURE_HOOK",
            "hooks": {"failure": "TASK_C", "success": "TASK_B"},
            "name": "SUCCESSANDFAILURE_HOOK",
            "project": "example",
            "tasks": ["TASK_A"],
        },
        {
            "fullname": "example.TASK_ROUTING",
            "name": "TASK_ROUTING",
            "project": "example",
            "queue": {"customs": {"TASK_B": "q2"}, "default": "q1"},
            "tasks": [
                "TASK_A",
                {"EXAMPLE_GROUP": {"tasks": ["TASK_B", "TASK_C"], "type": "group"}},
            ],
        },
        {
            "fullname": "example.WORKFLOW",
            "name": "WORKFLOW",
            "project": "example",
            "tasks": ["TASK_EXAMPLE"],
        },
        {
            "fullname": "schemas.SIMPLE_SCHEMA",
            "name": "SIMPLE_SCHEMA",
            "project": "schemas",
            "schema": {
                "properties": {"name": {"type": "string"}, "price": {"type": "number"}},
                "required": ["name"],
                "type": "object",
            },
            "tasks": ["TASK_EXAMPLE"],
        },
    ]



import importlib
import os
import pkgutil
from functools import partial
from pathlib import Path

from celery.schedules import crontab
from flask import Flask, Blueprint, jsonify, request, render_template
from werkzeug.exceptions import HTTPException

from director.api import api_bp
from director.extensions import cel, cel_workflows, db, schema, sentry, migrate
from director.settings import Config, UserConfig
from director.tasks.base import BaseTask
from director.utils import build_celery_schedule
from director.views import view_bp


with open(Path(__file__).parent.resolve() / "VERSION", encoding="utf-8") as version:
    __version__ = version.readline().rstrip()


# Proxify the task method
task = partial(cel.task, base=BaseTask)


# Provide the user config
config = UserConfig()


# Custom Flask class
class DirectorFlask(Flask):
    static_folder = "static"
    jinja_options = Flask.jinja_options.copy()
    jinja_options.update(
        dict(
            variable_start_string="%%",
            variable_end_string="%%",
        )
    )


# Create the application using a factory
def create_app(
    home_path=os.getenv("DIRECTOR_HOME"), config_path=os.getenv("DIRECTOR_CONFIG")
):
    app = DirectorFlask(__name__)
    c = Config(home_path, config_path)
    app.config.from_object(c)

    # Init User's config
    config.init()

    # Init Blueprints
    app.register_blueprint(api_bp)
    app.register_blueprint(view_bp)
    # Init the blueprint for the User's static assets
    app.register_blueprint(
        Blueprint(
            "user",
            __name__,
            static_url_path="/static/user",
            static_folder=str(Path(app.config["STATIC_FOLDER"])),
        )
    )

    # Error handler
    app.register_error_handler(HTTPException, lambda e: http_exception_handler(e))

    # Init extensions
    db.init_app(app)
    db.app = app

    migrate.init_app(
        app=app,
        db=db,
        directory=str(Path(__file__).resolve().parent / "migrations"),
    )
    schema.init_app(app)
    cel.init_app(app)
    cel_workflows.init_app(app)
    sentry.init_app(app)

    # Dict passed to the cleanup function
    retentions = {}

    # Register the periodic tasks for Celery Beat
    for workflow, conf in cel_workflows.workflows.items():
        retention = conf.get("retention", {}).get(
            "offset", app.config["DEFAULT_RETENTION_OFFSET"]
        )

        # A dict is built for the periodic cleaning if the retention is valid
        if retention >= 0:
            retentions[workflow] = retention

        if "periodic" in conf:
            periodic_conf = conf.get("periodic")
            periodic_payload = periodic_conf.get("payload", {})
            schedule_str, schedule_value = build_celery_schedule(
                workflow, periodic_conf
            )

            cel.conf.beat_schedule.update(
                {
                    f"periodic-{workflow}-{schedule_str}": {
                        "task": "director.tasks.periodic.execute",
                        "schedule": schedule_value,
                        "args": (
                            workflow,
                            periodic_payload,
                        ),
                    }
                }
            )

    if len(retentions):
        cel.conf.beat_schedule.update(
            {
                "periodic-cleanup": {
                    "task": "director.tasks.periodic.cleanup",
                    "schedule": crontab(minute=0, hour=0),
                    "args": (retentions,),
                }
            }
        )

    @app.teardown_request
    def session_clear(exception=None):
        db.session.remove()

    return app


def http_exception_handler(error):
    if request.path.startswith("/api"):
        return jsonify(error=str(error.description)), error.code

    return render_template("error.html", error=error), error.code


# Import director's submodules
def import_submodules(package, modules_to_import):
    """
    Import all submodules of a module, recursively, including subpackages.
    """
    if isinstance(package, str):
        package = importlib.import_module(package)
    results = {}

    for loader, full_name, is_pkg in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):
        name = full_name.split(".")[-1]
        if not name.startswith("_") and not name.startswith("tests"):

            if any(x in package.__name__ for x in modules_to_import) or any(
                x in name for x in modules_to_import
            ):
                results[full_name] = importlib.import_module(full_name)
                if is_pkg:
                    results.update(import_submodules(full_name, modules_to_import))
    return results


import_submodules(__name__, ("api", "models", "tasks", "views"))



import time

from celery.utils.log import get_task_logger
from celery import chain
from celery.utils import uuid


from director.extensions import cel
from director.models import StatusType
from director.models.workflows import Workflow
from director.models.tasks import Task


logger = get_task_logger(__name__)


@cel.task(name="celery.ping")
def ping():
    # type: () -> str
    """Simple task that just returns 'pong'."""
    return "pong"


@cel.task()
def start(workflow_id):
    logger.info(f"Opening the workflow {workflow_id}")
    workflow = Workflow.query.filter_by(id=workflow_id).first()

    workflow.status = StatusType.progress
    workflow.save()


@cel.task()
def end(workflow_id):
    # Waiting for the workflow status to be marked in error if a task failed
    time.sleep(0.5)

    logger.info(f"Closing the workflow {workflow_id}")
    workflow = Workflow.query.filter_by(id=workflow_id).first()

    if workflow.status != StatusType.error:
        workflow.status = StatusType.success
        workflow.save()


@cel.task()
def mark_as_canceled_pending_tasks(workflow_id):
    logger.info(f"Mark as cancelled pending tasks of the workflow {workflow_id}")
    tasks = Task.query.filter_by(workflow_id=workflow_id, status=StatusType.pending)
    for task in tasks:
        task.status = StatusType.canceled
        task.save()


@cel.task()
def failure_hooks_launcher(workflow_id, queue, tasks_names, payload):
    canvas = []

    for task_name in tasks_names:
        task_id = uuid()

        # We create the Celery task specifying its UID
        signature = cel.tasks.get(task_name).subtask(
            kwargs={"workflow_id": workflow_id, "payload": payload},
            task_id=task_id,
        )

        # Director task has the same UID
        task = Task(
            id=task_id,
            key=task_name,
            workflow_id=workflow_id,
            status=StatusType.pending,
            is_hook=True,
        )
        task.save()

        canvas.append(signature)

    canvas = chain(*canvas, task_id=uuid())

    result = canvas.apply_async()

    try:
        result.get()
    except Exception:
        pass

    task_id = uuid()
    signature_mark_as_canceled = cel.tasks.get(
        "director.tasks.workflows.mark_as_canceled_pending_tasks"
    ).subtask(
        args=(workflow_id,),
        task_id=task_id,
    )
    signature_mark_as_canceled.apply_async()



import os
import pytest

from werkzeug.security import generate_password_hash

from director.models.users import User


def test_auth_enabled_not_authorized(enable_auth, client):
    resp = client.get("/api/workflows")
    assert resp.status_code == 401
    assert "unauthorized" in resp.json["error"]


def test_auth_enabled_authorized(enable_auth, client):
    user = User(username="test_user", password=generate_password_hash("root"))
    user.save()

    # Enc test_user:root = dGVzdF91c2VyOnJvb3Q=
    resp = client.get(
        "/api/workflows", headers={"Authorization": "Basic dGVzdF91c2VyOnJvb3Q="}
    )
    assert resp.status_code == 200


def test_auth_disabled(enable_auth, client):
    resp = client.get("/api/ping")
    assert resp.status_code == 200
    assert resp.json == {"message": "pong"}



import os
import click

from urllib.parse import urlparse

from director.context import pass_ctx


@click.group()
def celery():
    """Celery commands"""


@celery.command(name="beat", context_settings=dict(ignore_unknown_options=True))
@click.option("--dev", "dev_mode", default=False, is_flag=True, type=bool)
@click.argument("beat_args", nargs=-1, type=click.UNPROCESSED)
def beat(dev_mode, beat_args):
    """Start the beat instance"""
    args = [
        "celery",
        "-A",
        "director._auto:cel",
        "beat",
    ]
    if dev_mode:
        args += [
            "--loglevel",
            "INFO",
        ]
    args += list(beat_args)
    os.execvp(args[0], args)


@celery.command("worker", context_settings=dict(ignore_unknown_options=True))
@click.option("--dev", "dev_mode", default=False, is_flag=True, type=bool)
@click.argument("worker_args", nargs=-1, type=click.UNPROCESSED)
def worker(dev_mode, worker_args):
    """Start a Celery worker instance"""
    args = [
        "celery",
        "-A",
        "director._auto:cel",
        "worker",
    ]
    if dev_mode:
        args += [
            "--loglevel",
            "INFO",
        ]
    args += list(worker_args)
    os.execvp(args[0], args)


@celery.command(name="flower", context_settings=dict(ignore_unknown_options=True))
@click.argument("flower_args", nargs=-1, type=click.UNPROCESSED)
@pass_ctx
def flower(ctx, flower_args):
    """Start the flower instance"""
    broker = ctx.app.config["CELERY_CONF"]["broker_url"]
    args = ["celery", "-b", broker, "flower"]
    args += list(flower_args)
    os.execvp(args[0], args)



from director.extensions import db
from director.models import BaseModel, StatusType
from director.models.utils import JSONBType


class Workflow(BaseModel):
    __tablename__ = "workflows"

    name = db.Column(db.String(255), nullable=False)
    project = db.Column(db.String(255), nullable=False)
    status = db.Column(db.Enum(StatusType), default=StatusType.pending, nullable=False)
    payload = db.Column(JSONBType, default={})
    periodic = db.Column(db.Boolean, default=False)

    def __str__(self):
        return f"{self.project}.{self.name}"

    def __repr__(self):
        return f"<Workflow {self.project}.{self.name}>"

    def to_dict(self, with_payload=True):
        d = super().to_dict()
        d.update(
            {
                "name": self.name,
                "project": self.project,
                "fullname": f"{self.project}.{self.name}",
                "status": self.status.value,
                "periodic": self.periodic,
            }
        )
        if with_payload:
            d["payload"] = self.payload
        return d



import os
from pathlib import Path

from environs import Env


HIDDEN_CONFIG = [
    "DIRECTOR_ENABLE_HISTORY_MODE",
    "DIRECTOR_REFRESH_INTERVAL",
    "DIRECTOR_API_URL",
    "DIRECTOR_FLOWER_URL",
    "DIRECTOR_DATABASE_URI",
    "DIRECTOR_DATABASE_POOL_RECYCLE",
    "DIRECTOR_BROKER_URI",
    "DIRECTOR_RESULT_BACKEND_URI",
    "DIRECTOR_SENTRY_DSN",
]


class Config(object):
    def __init__(self, home_path=None, config_path=None):
        if not home_path or not Path(home_path).resolve().exists():
            raise ValueError("environment variable DIRECTOR_HOME is not set correctly")
        self.DIRECTOR_HOME = env_path = str(home_path)

        if config_path:
            if not Path(config_path).resolve().exists():
                raise ValueError(
                    "environment variable DIRECTOR_CONFIG is not set correctly"
                )
            env_path = config_path

        env = Env()
        env.read_env(env_path)

        self.ENABLE_HISTORY_MODE = env.bool("DIRECTOR_ENABLE_HISTORY_MODE", False)
        self.ENABLE_CDN = env.bool("DIRECTOR_ENABLE_CDN", True)
        self.STATIC_FOLDER = env.str(
            "DIRECTOR_STATIC_FOLDER", str(Path(self.DIRECTOR_HOME).resolve() / "static")
        )
        self.API_URL = env.str("DIRECTOR_API_URL", "http://127.0.0.1:8000/api")
        self.FLOWER_URL = env.str("DIRECTOR_FLOWER_URL", "http://127.0.0.1:5555")
        self.WORKFLOWS_PER_PAGE = env.int("DIRECTOR_WORKFLOWS_PER_PAGE", 1000)
        self.REFRESH_INTERVAL = env.int("DIRECTOR_REFRESH_INTERVAL", 30000)
        self.REPO_LINK = env.str(
            "DIRECTOR_REPO_LINK", "https://github.com/ovh/celery-director"
        )
        self.DOCUMENTATION_LINK = env.str(
            "DIRECTOR_DOCUMENTATION_LINK", "https://ovh.github.io/celery-director"
        )

        # Authentication
        self.AUTH_ENABLED = env.bool("DIRECTOR_AUTH_ENABLED", False)

        # SQLAlchemy configuration
        self.SQLALCHEMY_TRACK_MODIFICATIONS = False
        self.SQLALCHEMY_DATABASE_URI = env.str("DIRECTOR_DATABASE_URI", "")
        self.SQLALCHEMY_ENGINE_OPTIONS = {
            "pool_recycle": env.int("DIRECTOR_DATABASE_POOL_RECYCLE", -1),
        }

        # Celery configuration
        self.CELERY_CONF = {
            "task_always_eager": False,
            "broker_url": env.str("DIRECTOR_BROKER_URI", "redis://localhost:6379/0"),
            "result_backend": env.str(
                "DIRECTOR_RESULT_BACKEND_URI", "redis://localhost:6379/1"
            ),
            "broker_transport_options": {"master_name": "director"},
        }

        # Sentry configuration
        self.SENTRY_DSN = env.str("DIRECTOR_SENTRY_DSN", "")

        # Default retention value (number of workflows to keep in the database)
        self.DEFAULT_RETENTION_OFFSET = env.int("DIRECTOR_DEFAULT_RETENTION_OFFSET", -1)

        # Enable Vue debug loading vue.js instead of vue.min.js
        self.VUE_DEBUG = env.bool("DIRECTOR_VUE_DEBUG", False)


class UserConfig(dict):
    """Handle the user configuration"""

    def init(self):
        envs = {
            k.split("DIRECTOR_")[1]: v
            for k, v in os.environ.items()
            if k.startswith("DIRECTOR_") and k not in HIDDEN_CONFIG
        }
        super().__init__(**envs)

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as e:
            raise AttributeError(f"Config '{e.args[0]}' not defined")



"""Add index on workflow_id in task table

Revision ID: 063ff371f2da
Revises: 3f8466b16023
Create Date: 2020-10-09 12:09:27.596429

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "063ff371f2da"
down_revision = "3f8466b16023"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        op.f("ix_tasks_workflow_id"), "tasks", ["workflow_id"], unique=False
    )


def downgrade():
    op.drop_index(op.f("ix_tasks_workflow_id"), table_name="tasks")



import imp
import json
from pathlib import Path
from json.decoder import JSONDecodeError

import yaml
import sentry_sdk
from celery import Celery
from flask_sqlalchemy import SQLAlchemy
from flask_json_schema import JsonSchema, JsonValidationError
from flask_migrate import Migrate
from pluginbase import PluginBase
from sqlalchemy.schema import MetaData
from sentry_sdk.integrations import celery as sentry_celery
from sentry_sdk.utils import capture_internal_exceptions
from celery.exceptions import SoftTimeLimitExceeded

from director.exceptions import SchemaNotFound, SchemaNotValid, WorkflowNotFound


class CeleryWorkflow:
    def __init__(self):
        self.app = None
        self.workflows = None

    def init_app(self, app):
        self.app = app
        self.path = Path(self.app.config["DIRECTOR_HOME"]).resolve() / "workflows.yml"
        with open(self.path) as f:
            self.workflows = yaml.load(f, Loader=yaml.SafeLoader)

        self.import_user_tasks()
        self.read_schemas()

    def get_by_name(self, name):
        workflow = self.workflows.get(name)
        if not workflow:
            raise WorkflowNotFound(f"Workflow {name} not found")
        return workflow

    def get_tasks(self, name):
        return self.get_by_name(name)["tasks"]

    def get_hook_task(self, name, hook_name):
        if (
            "hooks" in self.get_by_name(name)
            and hook_name in self.get_by_name(name)["hooks"]
        ):
            return self.get_by_name(name)["hooks"][hook_name]
        return None

    def get_failure_hook_task(self, name):
        return self.get_hook_task(name, "failure")

    def get_success_hook_task(self, name):
        return self.get_hook_task(name, "success")

    def get_queue(self, name):
        try:
            return self.get_by_name(name)["queue"]
        except KeyError:
            return "celery"

    def import_user_tasks(self):
        self.plugin_base = PluginBase(package="director.foobar")

        folder = Path(self.app.config["DIRECTOR_HOME"]).resolve()
        self.plugin_source = self.plugin_base.make_plugin_source(
            searchpath=[str(folder)]
        )

        tasks = Path(folder / "tasks").glob("**/*.py")
        with self.plugin_source:
            for task in tasks:
                if task.stem == "__init__":
                    continue

                name = str(task.relative_to(folder))[:-3].replace("/", ".")
                __import__(
                    self.plugin_source.base.package + "." + name,
                    globals(),
                    {},
                    ["__name__"],
                )

    def read_schemas(self):
        folder = Path(self.app.config["DIRECTOR_HOME"]).resolve()

        for name, conf in self.workflows.items():
            if "schema" in conf:
                path = Path(folder / "schemas" / f"{conf['schema']}.json")

                try:
                    schema = json.loads(open(path).read())
                except FileNotFoundError:
                    raise SchemaNotFound(
                        f"Schema '{conf['schema']}' not found ({path})"
                    )
                except JSONDecodeError as e:
                    raise SchemaNotValid(f"Schema '{conf['schema']}' not valid ({e})")

                self.workflows[name]["schema"] = schema


# Celery Extension
class FlaskCelery(Celery):
    def __init__(self, *args, **kwargs):
        kwargs["include"] = ["director.tasks"]
        super(FlaskCelery, self).__init__(*args, **kwargs)

        if "app" in kwargs:
            self.init_app(kwargs["app"])

    def init_app(self, app):
        self.app = app
        self.conf.update(app.config.get("CELERY_CONF", {}))


# Sentry Extension
class DirectorSentry:
    def __init__(self):
        self.app = None

    def init_app(self, app):
        self.app = app

        if self.app.config["SENTRY_DSN"]:
            sentry_celery._make_event_processor = self.custom_event_processor
            sentry_sdk.init(
                dsn=self.app.config["SENTRY_DSN"],
                integrations=[sentry_celery.CeleryIntegration()],
            )

    def enrich_tags(self, tags, workflow_id, task):
        from director.models.workflows import Workflow

        with self.app.app_context():
            workflow_obj = Workflow.query.filter_by(id=workflow_id).first()
            workflow = {
                "id": str(workflow_obj.id),
                "project": workflow_obj.project,
                "name": str(workflow_obj),
            }

        tags.update(
            {
                "celery_task_name": task.name,
                "director_workflow_id": workflow.get("id"),
                "director_workflow_project": workflow.get("project"),
                "director_workflow_name": workflow.get("name"),
            }
        )
        return tags

    def enrich_extra(self, extra, args, kwargs):
        extra.update({"workflow-payload": kwargs["payload"], "task-args": args})
        return extra

    def custom_event_processor(self, task, uuid, args, kwargs, request=None):
        """
        This function is the same as the original, except that we
        add custom tags and extras about the workflow object.

        Published under a BSD-2 license and available at:
        https://github.com/getsentry/sentry-python/blob/0.16.3/sentry_sdk/integrations/celery.py#L176
        """

        def event_processor(event, hint):
            with capture_internal_exceptions():
                tags = event.setdefault("tags", {})
                tags["celery_task_id"] = uuid
                extra = event.setdefault("extra", {})
                extra["celery-job"] = {
                    "task_name": task.name,
                    "args": args,
                    "kwargs": kwargs,
                }

                # Director custom fields (references are used by Sentry,
                # no need to retrieve the new values)
                self.enrich_tags(tags, kwargs["workflow_id"], task)
                self.enrich_extra(extra, args, kwargs)

            if "exc_info" in hint:
                with capture_internal_exceptions():
                    if issubclass(hint["exc_info"][0], SoftTimeLimitExceeded):
                        event["fingerprint"] = [
                            "celery",
                            "SoftTimeLimitExceeded",
                            getattr(task, "name", task),
                        ]

            return event

        return event_processor


# List of extensions
db = SQLAlchemy(
    metadata=MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",
            "uq": "uq_%(table_name)s_%(column_0_name)s",
            "ck": "ck_%(table_name)s_%(column_0_name)s",
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
            "pk": "pk_%(table_name)s",
        }
    )
)
migrate = Migrate()
schema = JsonSchema()
cel = FlaskCelery("director")
cel_workflows = CeleryWorkflow()
sentry = DirectorSentry()



from pathlib import Path

from setuptools import find_packages, setup

with open(
    Path(__file__).parent.resolve() / "director" / "VERSION", encoding="utf-8"
) as ver:
    version = ver.readline().rstrip()

with open("requirements.txt", encoding="utf-8") as req:
    requirements = [r.rstrip() for r in req.readlines()]

long_description = ""
try:
    with open("README.md", encoding="utf-8") as readme:
        long_description = readme.read()
except FileNotFoundError:
    pass

dev_requirements = [
    "tox==3.5.3",
    "black==22.3.0",
]

doc_requirements = ["mkdocs==1.3.0", "mkdocs-material==8.2.9"]


setup(
    name="celery-director",
    version=version,
    description="Celery Director",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="BSD",
    author="OVHcloud",
    author_email="opensource@ovhcloud.com",
    url="https://github.com/ovh/celery-director",
    packages=find_packages(),
    install_requires=requirements,
    extras_require={
        "dev": dev_requirements,
        "ci": ["pytest", "pytest-cov"],
        "doc": doc_requirements,
    },
    include_package_data=True,
    entry_points={"console_scripts": ["director=director.cli:cli"],},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: System :: Monitoring",
    ],
    python_requires="~=3.7",
)



import time

import pytest
from celery.result import GroupResult
from celery.schedules import crontab
from kombu.exceptions import EncodeError

from director import build_celery_schedule
from director.exceptions import WorkflowSyntaxError
from director.models.tasks import Task
from director.models.workflows import Workflow
from director.builder import WorkflowBuilder


KEYS = ["id", "created", "updated", "task"]


def test_execute_one_task_success(app, create_builder):
    workflow, builder = create_builder("example", "WORKFLOW", {})
    assert workflow["status"] == "pending"

    # Canvas has been built
    assert len(builder.canvas) == 3
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_EXAMPLE"

    # Tasks added in DB
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 1
    assert tasks[0].key == "TASK_EXAMPLE"
    assert tasks[0].status.value == "pending"

    # Tasks executed in Celery
    result = builder.run()
    assert result.get() is None
    assert result.parent.parent.get() is None
    assert result.parent.get() == "task_example"
    assert result.parent.state == "SUCCESS"

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        task = Task.query.filter_by(id=tasks[0].id).first()
        workflow = Workflow.query.filter_by(id=task.workflow_id).first()
    assert workflow.status.value == "success"
    assert task.status.value == "success"


def test_execute_one_task_error(app, create_builder):
    workflow, builder = create_builder("example", "ERROR", {})
    assert workflow["status"] == "pending"

    # Canvas has been built
    assert len(builder.canvas) == 3
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_ERROR"

    # Tasks added in DB
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 1
    assert tasks[0].key == "TASK_ERROR"
    assert tasks[0].status.value == "pending"

    # Tasks executed in Celery
    result = builder.run()
    with pytest.raises(ZeroDivisionError):
        assert result.get()

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        task = Task.query.filter_by(id=tasks[0].id).first()
        workflow = Workflow.query.filter_by(id=task.workflow_id).first()
    assert workflow.status.value == "error"
    assert task.status.value == "error"


def test_execute_chain_success(app, create_builder):
    workflow, builder = create_builder("example", "SIMPLE_CHAIN", {})
    assert workflow["status"] == "pending"

    # Canvas has been built
    assert len(builder.canvas) == 5
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert [c.task for c in builder.canvas[1:-1]] == ["TASK_A", "TASK_B", "TASK_C"]

    # Tasks added in DB
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 3
    assert [n.key for n in tasks] == ["TASK_A", "TASK_B", "TASK_C"]
    assert set([n.status.value for n in tasks]) == {
        "pending",
    }

    # Tasks executed in Celery
    result = builder.run()
    assert result.get() is None
    assert result.parent.parent.parent.parent.get() is None
    assert result.parent.get() == "task_c"
    assert result.parent.state == "SUCCESS"
    assert result.parent.parent.get() == "task_b"
    assert result.parent.parent.state == "SUCCESS"
    assert result.parent.parent.parent.get() == "task_a"
    assert result.parent.parent.parent.state == "SUCCESS"

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        tasks = Task.query.filter_by(id=tasks[0].id).all()
        workflow = Workflow.query.filter_by(id=tasks[0].workflow_id).first()
    assert workflow.status.value == "success"

    for task in tasks:
        assert task.status.value == "success"


def test_execute_chain_error(app, create_builder):
    workflow, builder = create_builder("example", "SIMPLE_CHAIN_ERROR", {})
    assert workflow["status"] == "pending"

    # Canvas has been built
    assert len(builder.canvas) == 5
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert [c.task for c in builder.canvas[1:-1]] == ["TASK_A", "TASK_B", "TASK_ERROR"]

    # Tasks added in DB
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 3
    assert [n.key for n in tasks] == ["TASK_A", "TASK_B", "TASK_ERROR"]
    assert set([n.status.value for n in tasks]) == {
        "pending",
    }

    # Tasks executed in Celery
    result = builder.run()
    with pytest.raises(ZeroDivisionError):
        assert result.get()

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        task_a = Task.query.filter_by(key="TASK_A").first()
        task_b = Task.query.filter_by(key="TASK_B").first()
        task_error = Task.query.filter_by(key="TASK_ERROR").first()
        workflow = Workflow.query.filter_by(id=task_a.workflow_id).first()
    assert task_a.status.value == "success"
    assert task_b.status.value == "success"
    assert task_error.status.value == "error"
    assert workflow.status.value == "error"


def test_execute_group_success(app, create_builder):
    workflow, builder = create_builder("example", "SIMPLE_GROUP", {})
    assert workflow["status"] == "pending"

    # Canvas has been built
    assert len(builder.canvas) == 4
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_A"
    group_tasks = builder.canvas[2].tasks
    assert len(group_tasks) == 2
    assert [group_tasks[0].task, group_tasks[1].task] == [
        "TASK_B",
        "TASK_C",
    ]

    # Tasks added in DB
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 3
    assert [n.key for n in tasks] == ["TASK_A", "TASK_B", "TASK_C"]
    assert set([n.status.value for n in tasks]) == {
        "pending",
    }

    # Tasks executed in Celery
    result = builder.run()
    assert result.get() is None
    assert result.parent.parent.get() == "task_a"
    assert isinstance(result.parent, GroupResult)
    assert result.parent.get() == ["task_b", "task_c"]

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        tasks = Task.query.filter_by(id=tasks[0].id).all()
        workflow = Workflow.query.filter_by(id=tasks[0].workflow_id).first()
    assert workflow.status.value == "success"

    for task in tasks:
        assert task.status.value == "success"


def test_execute_group_error(app, create_builder):
    workflow, builder = create_builder("example", "SIMPLE_GROUP_ERROR", {})
    assert workflow["status"] == "pending"

    # Canvas has been built
    assert len(builder.canvas) == 4
    assert builder.canvas[0].task == "director.tasks.workflows.start"
    assert builder.canvas[-1].task == "director.tasks.workflows.end"
    assert builder.canvas[1].task == "TASK_A"
    group_tasks = builder.canvas[2].tasks
    assert len(group_tasks) == 2
    assert [group_tasks[0].task, group_tasks[1].task] == ["TASK_ERROR", "TASK_C"]

    # Tasks added in DB
    with app.app_context():
        tasks = Task.query.order_by(Task.created_at.asc()).all()
    assert len(tasks) == 3
    assert [n.key for n in tasks] == ["TASK_A", "TASK_ERROR", "TASK_C"]
    assert set([n.status.value for n in tasks]) == {
        "pending",
    }

    # Tasks executed in Celery
    result = builder.run()
    with pytest.raises(ZeroDivisionError):
        assert result.get()

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        task_a = Task.query.filter_by(key="TASK_A").first()
        task_error = Task.query.filter_by(key="TASK_ERROR").first()
        task_c = Task.query.filter_by(key="TASK_C").first()
        workflow = Workflow.query.filter_by(id=task_a.workflow_id).first()
    assert task_a.status.value == "success"
    assert task_error.status.value == "error"
    assert task_c.status.value == "success"
    assert workflow.status.value == "error"


@pytest.mark.skip_no_worker()
def test_execute_celery_error_one_task(app, create_builder):
    workflow, builder = create_builder("example", "CELERY_ERROR_ONE_TASK", {})
    assert workflow["status"] == "pending"

    # Tasks executed in Celery
    result = builder.run()
    with pytest.raises(EncodeError):
        assert result.get()

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        task = Task.query.order_by(Task.created_at.asc()).first()
        workflow = Workflow.query.filter_by(id=task.workflow_id).first()
    assert workflow.status.value == "error"
    assert task.status.value == "error"


@pytest.mark.skip_no_worker()
def test_execute_celery_error_multiple_tasks(app, create_builder):
    workflow, builder = create_builder("example", "CELERY_ERROR_MULTIPLE_TASKS", {})
    assert workflow["status"] == "pending"

    # Tasks executed in Celery
    result = builder.run()
    with pytest.raises(EncodeError):
        assert result.get()

    # DB rows status updated
    time.sleep(0.5)
    with app.app_context():
        task_a = Task.query.filter_by(key="TASK_A").first()
        task_celery_error = Task.query.filter_by(key="TASK_CELERY_ERROR").first()
        workflow = Workflow.query.filter_by(id=task_a.workflow_id).first()
    assert task_a.status.value == "success"
    assert task_celery_error.status.value == "error"
    assert workflow.status.value == "error"


@pytest.mark.skip_no_worker()
def test_cancel_workflow(app):

    with app.app_context():
        workflow = Workflow(
            project="example", name="DELAY_TASK", payload={}, periodic=False
        )
        workflow.save()
        builder = WorkflowBuilder(workflow.id)
        builder.build()
        assert workflow.status.value == "pending"

        # Tasks executed in Celery
        builder.run()

        # DB rows status updated
        time.sleep(1)

        workflow = Workflow.query.filter_by(id=workflow.id).first()

        assert workflow.tasks[0].status.value == "progress"

        builder.cancel()

        # DB rows status updated
        time.sleep(0.5)

        workflow = Workflow.query.filter_by(id=workflow.id).first()

        assert workflow.status.value == "canceled"
        assert workflow.tasks[0].status.value == "canceled"

        # Wait for the restart of the only prefork
        time.sleep(1)


def test_return_values(app, create_builder):
    workflow, builder = create_builder("example", "RETURN_VALUES", {})
    builder.run()

    time.sleep(0.5)

    with app.app_context():
        tasks = {t.key: t.result for t in Task.query.all()}

    assert tasks["STR"] == "return_value"
    assert tasks["INT"] == 1234
    assert tasks["LIST"] == ["jack", "sape", "guido"]
    assert tasks["NONE"] is None
    assert tasks["DICT"] == {"foo": "bar"}
    assert tasks["NESTED"] == {
        "jack": 4098,
        "sape": 4139,
        "guido": 4127,
        "nested": {"foo": "bar"},
        "none": None,
        "list": ["jack", "sape", "guido"],
    }


@pytest.mark.skip_no_worker()
def test_execute_workflow_withfailurehook(app):

    with app.app_context():
        workflow = Workflow(
            project="example", name="FAILURE_HOOK", payload={}, periodic=False
        )
        workflow.save()
        workflow_id = workflow.id
        builder = WorkflowBuilder(workflow.id)
        builder.build()
        assert workflow.status.value == "pending"

        # Tasks executed in Celery
        result = builder.run()
        with pytest.raises(ZeroDivisionError):
            assert result.get()

    # DB rows status updated
    time.sleep(0.5)

    with app.app_context():
        workflow = Workflow.query.filter_by(id=workflow_id).first()

        assert workflow.status.value == "error"
        assert workflow.tasks[0].status.value == "success"
        assert workflow.tasks[1].status.value == "error"
        assert workflow.tasks[2].status.value == "success"


@pytest.mark.skip_no_worker()
def test_execute_workflow_withsuccesshook(app):

    with app.app_context():
        workflow = Workflow(
            project="example", name="SUCCESSANDFAILURE_HOOK", payload={}, periodic=False
        )
        workflow.save()
        workflow_id = workflow.id
        builder = WorkflowBuilder(workflow.id)
        builder.build()
        assert workflow.status.value == "pending"

        # Tasks executed in Celery
        result = builder.run()
        assert result.get() == None

    # DB rows status updated
    time.sleep(0.5)

    with app.app_context():
        workflow = Workflow.query.filter_by(id=workflow_id).first()

        assert workflow.status.value == "success"
        assert workflow.tasks[0].status.value == "success"
        assert workflow.tasks[1].status.value == "success"


def test_return_exception(app, create_builder):
    workflow, builder = create_builder("example", "RETURN_EXCEPTION", {})
    builder.run()

    time.sleep(0.5)
    with app.app_context():
        tasks = {t.key: t.result for t in Task.query.all()}

    assert tasks["STR"] == "return_value"
    assert list(tasks["TASK_ERROR"].keys()) == ["exception", "traceback"]
    assert tasks["TASK_ERROR"]["exception"] == "division by zero"
    assert tasks["TASK_ERROR"]["traceback"].startswith(
        "Traceback (most recent call last)"
    )
    assert "ZeroDivisionError: division by zero" in tasks["TASK_ERROR"]["traceback"]


def test_build_celery_schedule_float_with_payload():
    float_schedule = {"payload": {}, "schedule": 30.0}
    assert ("30.0", 30.0) == build_celery_schedule(
        "workflow_schedule_float", float_schedule
    )


def test_build_celery_schedule_float():
    float_schedule = {"schedule": 30.0}
    assert ("30.0", 30.0) == build_celery_schedule(
        "workflow_schedule_float", float_schedule
    )


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (
            "1 * * * *",
            crontab(
                minute="1",
                hour="*",
                day_of_week="*",
                day_of_month="*",
                month_of_year="*",
            ),
        ),
        (
            "* 1 * * *",
            crontab(
                minute="*",
                hour="1",
                day_of_week="*",
                day_of_month="*",
                month_of_year="*",
            ),
        ),
        (
            "* * 1 * *",
            crontab(
                minute="*",
                hour="*",
                day_of_week="1",
                day_of_month="*",
                month_of_year="*",
            ),
        ),
        (
            "* * * 1 *",
            crontab(
                minute="*",
                hour="*",
                day_of_week="*",
                day_of_month="1",
                month_of_year="*",
            ),
        ),
        (
            "* * * * 1",
            crontab(
                minute="*",
                hour="*",
                day_of_week="*",
                day_of_month="*",
                month_of_year="1",
            ),
        ),
        (
            "*/10 */11 */12 */13 */14",
            crontab(
                minute="*/10",
                hour="*/11",
                day_of_week="*/12",
                day_of_month="*/13",
                month_of_year="*/14",
            ),
        ),
    ],
)
def test_build_celery_schedule_crontab(test_input, expected):
    cron_schedule = {"schedule": test_input}
    assert (test_input, expected) == build_celery_schedule(
        "workflow_crontab", cron_schedule
    )


def test_build_celery_interval():
    float_schedule = {"interval": 30.0}
    assert ("30.0", 30.0) == build_celery_schedule(
        "workflow_schedule_float", float_schedule
    )


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (
            "1 * * * *",
            crontab(
                minute="1",
                hour="*",
                day_of_month="*",
                month_of_year="*",
                day_of_week="*",
            ),
        ),
        (
            "* 1 * * *",
            crontab(
                minute="*",
                hour="1",
                day_of_month="*",
                month_of_year="*",
                day_of_week="*",
            ),
        ),
        (
            "* * 1 * *",
            crontab(
                minute="*",
                hour="*",
                day_of_month="1",
                month_of_year="*",
                day_of_week="*",
            ),
        ),
        (
            "* * * 1 *",
            crontab(
                minute="*",
                hour="*",
                day_of_month="*",
                month_of_year="1",
                day_of_week="*",
            ),
        ),
        (
            "* * * * 1",
            crontab(
                minute="*",
                hour="*",
                day_of_month="*",
                month_of_year="*",
                day_of_week="1",
            ),
        ),
        (
            "*/10 */11 */12 */13 */14",
            crontab(
                minute="*/10",
                hour="*/11",
                day_of_month="*/12",
                month_of_year="*/13",
                day_of_week="*/14",
            ),
        ),
    ],
)
def test_build_celery_crontab(test_input, expected):
    cron_schedule = {"crontab": test_input}
    assert (test_input, expected) == build_celery_schedule(
        "workflow_crontab", cron_schedule
    )


def test_build_celery_invalid_crontab():
    # missing one element on the crontab syntax
    periodic_conf = {"crontab": "* * * *"}
    with pytest.raises(WorkflowSyntaxError):
        build_celery_schedule("workflow_invalid_crontab", periodic_conf)


def test_build_celery_invalid_schedule():
    cron_schedule = {"crontab": "* * * * 12"}
    with pytest.raises(WorkflowSyntaxError):
        build_celery_schedule("workflow_invalid_crontab", cron_schedule)


def test_build_celery_invalid_periodic_key():
    cron_schedule = {"non_valid_key": "* * * * *"}
    with pytest.raises(WorkflowSyntaxError):
        build_celery_schedule("workflow_invalid_key", cron_schedule)
